/* DD2CPCDISC v0.1 
 * Thomas Kjaer, Apr/Jun - 1994
 *
 * DDTRANS is a special Un*x version of 
 *   CPCTRANS v1.1
 *   Marco Vieth, 9.12.1993 (6.12.1993)
 *   transfer CPC-disks to CPCEMU disk-images and vice versa
 *
 * Used with permission from the author of CPCTRANS
 *
 * The program works on exact disc-images made with the Un*x command 'dd',
 * which hopefully for you, supports DDDSDT disks. 
 * Only DD disc, e.g. vortex or PCW.
 *
 * to make a 'dd' file from a DD disk do this:
 *   dd if=/dev/fd0 of=image bs=737280
 *
 * Start this program and specify 'image' as the 'dd' file
 *
 * to write a DD disc from a 'dd' file do this:
 *   dd if=image of=/dev/fd0 bs=737280
 *
 * This program will also make newly formatted discs for the Amstrad emulator
 * CPCEMU by Marco Vieth, and the Sinclair Spectrum +3 emulator xzx for
 * Un*x/X-windows by Des Herriott.
 *
 * developed under Linux 1.0 with gcc
 */

static char rcsid[] = "$Id: ddtrans.c,v 1.3 1997/01/03 09:55:07 erik Rel $";
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#define MAXSEC			18
#define DRV_READ		2
#define DRV_WRITE		3
#define IMG_DISK_INFO	0x100
#define IMG_TRACK_INFO	0x100
typedef unsigned char byte;
typedef unsigned short int  word;
typedef struct {
	byte tracks;
	byte heads;
	word tsize;

} Disk_info;
typedef struct {
	byte track;
	byte head;
	byte sector;
	byte bps;
	byte state1;
	byte state2;
	byte unused1;
	byte unused2;
} Sector_info;
typedef struct {
	byte track;
	byte head;

	byte bps;
	byte spt;
	byte gap3;
	byte fill;

	Sector_info si[MAXSEC];

	byte snum;
	byte fsec;
} Track_info;
typedef struct {
	FILE *fp;
	char *fn;
} Drive;
typedef struct {
	char fname[80];
	FILE *f;
	byte disk_i[IMG_DISK_INFO];
	byte track_i[IMG_TRACK_INFO];
} Image;
static byte sbuff[0x1000];
static byte dpb_buff[0x10];

static char dpb_p3[] = {
	0x00, 0x00, 0x28, 0x09, 0x02, 0x01, 0x03, 0x02,
	0x2a, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};
static char dpb_data[] = {
	0x02, 0x00, 0x28, 0x09, 0x02, 0x00, 0x03, 0x02,
	0x2a, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};
static char dpb_system[] = {
	0x01, 0x00, 0x28, 0x09, 0x02, 0x02, 0x03, 0x02,
	0x2a, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};
static char dpb_pcw[] = {
	0x03, 0x01, 0x50, 0x09, 0x02, 0x01, 0x04, 0x04,
	0x2a, 0x52, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};





static char
upcase(char ch)
{
	return (((ch>='a') && (ch<='z')) ? (ch - 0x20) : ch);
}
static char
*input(void)
{
	static char line[80];

	gets(line);
	return line;
}





static int
drv_rw_sector(byte command, Drive *dv, Track_info *t, byte *buffer)
{
#ifdef DEBUG
	printf("command=%d, drive=%s, head=%02d, track=%02d, sector=%02X\n",
		   command, dv->fn, t->head, t->track, (t->si[t->snum]).sector);
#endif
	if (command == DRV_READ) {
		if (fseek(dv->fp, 4608 * (2 * t->track + t->head) + 512 *
				  ((t->si[t->snum]).sector - 1), SEEK_SET)){
			fprintf(stderr, "error in seek\n");
			fclose(dv->fp);
			exit(1);
		}
		if (fread(sbuff, 512, 1, dv->fp) != 1) {
			fprintf(stderr, "error in read\n");
			fclose(dv->fp);
			exit(1);
		}
	}
	if (command == DRV_WRITE) {
		if (fseek(dv->fp, 4608 * (2 * t->track + t->head) + 512 *
			((t->si[t->snum]).sector - 1) , SEEK_SET)){
			fprintf(stderr, "error in seek\n");
			fclose(dv->fp);
			exit(1);
		}
		if (fwrite(sbuff, 512, 1, dv->fp) != 1) {
			fprintf(stderr, "error in write\n");
			fclose(dv->fp);
			exit(1);
		}
		}
#ifdef DEBUG
	printf("ok.\n");
#endif
	return 0;
}
static int
drv_read_sector(Drive *dv, Track_info *t, byte *buffer)
{
	return ( drv_rw_sector(DRV_READ, dv, t, buffer) );
}
static int
drv_write_sector(Drive *dv, Track_info *t, byte *buffer)
{
  return ( drv_rw_sector(DRV_WRITE, dv, t, buffer) );
}





static void
create_disk_info(byte *s, Disk_info *d)
{
	static char ident[]  = "MV - CPCEMU Disk-File\r\nDisk-Info\r\n";
	memset(s, 0, IMG_DISK_INFO);
	strcpy((char *)s, ident);
	s += 0x30;
	*(s++) = d->tracks;
	*(s++) = d->heads;
	*(s++) = (byte)d->tsize;
	*(s++) = (d->tsize >> 8);
}
static void
create_track_info(byte *s, Track_info *t)
{
	static char ident[] = "Track-Info\r\n";
	byte sec;
	memset(s, 0, IMG_TRACK_INFO);
	strcpy((char *)s, ident);
	s += 0x10;
	*(s++) = t->track;
	*(s++) = t->head;
	*(s++) = 0;
	*(s++) = 0;

	*(s++) = t->bps;
	*(s++) = t->spt;
	*(s++) = t->gap3;
	*(s++) = t->fill;

	for (sec = 0; sec < t->spt; sec++) {
		*(s++) = t->track;
		*(s++) = t->head;
		*(s++) = sec + t->fsec;
		*(s++) = t->bps;
		*(s++) = 0;
		*(s++) = 0;
		*(s++) = 0;
		*(s++) = 0;
	}
}
static int
img_write(FILE *stream, byte *buf, long len)
{
	if (fwrite(buf, sizeof(byte), len, stream) != len) {
		fprintf(stderr, "error fwrite()\n");
		return 1;
	}
	return 0;
}
static int
drv_to_img(Drive *dv, Disk_info *d, Track_info *t, Image *im)
{
	byte trk, hd, sec;
	create_disk_info(im->disk_i, d);
	if (img_write(im->f, im->disk_i, IMG_DISK_INFO))  return 1;
	for (trk = 0; trk < d->tracks; trk++) {
		t->track = trk;
		for (hd = 0; hd < d->heads; hd++) {
			t->head = hd;
			printf("\rLoading track %d  head %d ...  ", trk, hd);
			create_track_info(im->track_i, t);
			if (img_write(im->f, im->track_i, IMG_TRACK_INFO))
				return 1;
			for (sec = 0; sec < t->spt; sec++) {
				t->snum = sec;
				t->si[t->snum].sector = sec + t->fsec;
				if (drv_read_sector(dv, t, sbuff))
					return 1;
				if (img_write(im->f, sbuff, (0x0080 << t->bps)))
					return 1;
			}
		}
	}
	return 0;
}
static int
format_to_img(Drive *dv, Disk_info *d, Track_info *t, Image *im)
{
	byte trk, hd, sec;
	create_disk_info(im->disk_i, d);
	if (img_write(im->f, im->disk_i, IMG_DISK_INFO))
		return 1;
	for (trk = 0; trk < d->tracks; trk++) {
		t->track = trk;
		for (hd = 0; hd < d->heads; hd++) {
			t->head = hd;
			printf("\rFormatting track %d  head %d ...  ", trk, hd);
			create_track_info(im->track_i, t);
			if (img_write(im->f, im->track_i, IMG_TRACK_INFO))
				return 1;
			for (sec = 0; sec < t->spt; sec++) {
				t->snum = sec;
				t->si[t->snum].sector = sec + t->fsec;

				memset(sbuff, t->fill, 0x0080 << t->bps);
				if (trk == 0 && hd == 0 && sec == 0)
					memcpy(sbuff, dpb_buff, 0x10);
				if (img_write(im->f, sbuff, (0x0080 << t->bps)))
					return 1;
			}
		}
	}
	return 0;
}





static int
extract_disk_info(byte *s, Disk_info *d)
{
	static char ident[]  = "MV - CPCEMU Disk-File\r\nDisk-Info\r\n";
	if (memcmp(s, ident, 8))
		return 1;
	s += 0x30;
	d->tracks = *(s++);
	d->heads = *(s++);
	d->tsize = *(s++);
	d->tsize |= (*(s++) << 8);
	return 0;
}
static int
extract_track_info(byte *s, Track_info *t)
{
	static char ident[] = "Track-Info\r\n";
#ifdef DEBUG
	printf("extract_track_info\n");
#endif
	if (memcmp(s, ident, 5))
		return 1;
	s += 0x10;
	t->track = *(s++);
	t->head = *(s++);
	s+=2;

	t->bps = *(s++);
	t->spt = *(s++);
	t->gap3 = *(s++);
	t->fill = *(s++);

	memcpy(t->si, s, t->spt * sizeof(Sector_info));
	t->fsec = t->si[0].sector;
#ifdef DEBUG
	printf("t->fsec = %02X\n", t->fsec);
#endif
	return 0;
}
static int
img_read(FILE *stream, byte *buf, long len)
{
	if (fread(buf, sizeof(byte), len, stream) != len) {
		fprintf(stderr, "error fread()\n");
		return 1;
	}
	return 0;
}
static int
img_to_drv(Drive *dv, Disk_info *d, Track_info *t, Image *im)
{
	byte trk, hd, sec;
	if (img_read(im->f, im->disk_i, IMG_DISK_INFO))  return 1;
	if (extract_disk_info(im->disk_i, d))  return 1;
	for (trk = 0; trk < d->tracks; trk++) {
		for (hd = 0; hd < d->heads; hd++) {
			printf("\rSaving track %d  head %d ...  ", trk, hd);
			if (img_read(im->f, im->track_i, IMG_TRACK_INFO))
				return 1;
			if (extract_track_info(im->track_i, t))
				return 1;
			if ((trk != t->track) || (hd != t->head))
				return 1;
			for (sec = 0; sec < t->spt; sec++) {
				t->snum = sec;
				if (img_read(im->f, sbuff, (0x0080 << t->bps)))
					return 1;
				if (drv_write_sector(dv, t, sbuff))
					return 1;
			}
		}
	}
	return 0;
}





static void
set_f_data(Disk_info *d, Track_info *t)
{
	d->tracks = 40;
	d->heads = 1;
	t->bps = 0x02;
	t->spt = 9;
	t->gap3 = 0x4E;
	t->fill = 0xE5;
	t->fsec = 0xC1;
}
static void
set_f_vortex(Disk_info *d, Track_info *t)
{
	d->tracks = 80;
	d->heads = 2;
	t->bps = 0x02;
	t->spt = 9;
	t->gap3 = 0x4E;
	t->fill = 0xE5;
	t->fsec = 0x01;
}

static void
select_format_form(Disk_info *d, Track_info *t)
{
	char *s;
	d->tsize = 0xffff;
	t->track = 0xff;
	t->head = 0xff;
	t->snum = 0xff;
	set_f_data(d, t);
	memcpy(dpb_buff, dpb_data, 0x10);
	printf("Select format :\n");
	do {
		printf("d)ata 9x512x40x1; s)ystem 9x512x40x1; v)ortex 9x512x80x2;\n");
		printf("+)3 dos 9x512x40x1; a)mstrad PCW 9x512x80x2;\n");
		printf("m)odify; o)k\n");
		printf("Your choice: ");
		s = input();
		switch (upcase(s[0])) {
			case 'D' :
				set_f_data(d, t);
				memcpy(dpb_buff, dpb_data, 0x10);
				printf("DATA-format set.\n");
				break;
			case 'S' :
				set_f_data(d, t);
				t->fsec = 0x41;
				memcpy(dpb_buff, dpb_system, 0x10);
				printf("SYSTEM-format set.\n");
				break;
			case 'V' :
				set_f_vortex(d, t);
				memcpy(dpb_buff, dpb_pcw, 0x10);
				printf("VORTEX-format set.\n");
				break;
			case '+' :
				set_f_data(d, t);
				t->fsec = 0x00;
				memcpy(dpb_buff, dpb_p3, 0x10);
				printf("+3 DOS-format set.\n");
				break;
			case 'A' :
				set_f_vortex(d, t);
				memcpy(dpb_buff, dpb_pcw, 0x10);
				printf("Amstrad PCW-format set.\n");
				break;
			case 'M' :
				printf("Modify: not implemented.\n");
				break;
			case 'O' :
				printf("Ok.\n");
				break;
			default :
				printf("unknown format.\n");
		}
	} while (upcase(s[0]) != 'O');
	d->tsize = t->spt * (0x0080 << t->bps) + IMG_DISK_INFO;
}
static void
select_format(Disk_info *d, Track_info *t)
{
	char *s;
	d->tsize = 0xffff;
	t->track = 0xff;
	t->head = 0xff;
	t->snum = 0xff;
	set_f_vortex(d, t);
	memcpy(dpb_buff, dpb_pcw, 0x10);
	printf("Select format :\n");
	do {
		printf("v)ortex 9x512x80x2; a)mstrad PCW 9x512x80x2;\n");
		printf("m)odify; o)k\n");
		printf("Your choice: ");
		s = input();
		switch (upcase(s[0])) {
			case 'V' :
				set_f_vortex(d, t);
				memcpy(dpb_buff, dpb_pcw, 0x10);
				printf("VORTEX-format set.\n");
				break;
			case 'A' :
				set_f_vortex(d, t);
				memcpy(dpb_buff, dpb_pcw, 0x10);
				printf("Amstrad PCW-format set.\n");
				break;
			case 'M' :
				printf("Modify: not implemented.\n");
				break;
			case 'O' :
				printf("Ok.\n");
				break;
			default :
				printf("unknown format.\n");
		}
	} while (upcase(s[0]) != 'O');
	d->tsize = t->spt * (0x0080 << t->bps) + IMG_DISK_INFO;
}





static void
set_drive(Drive *drv)
{
	char *s;
	printf("filename of 'dd' image: ");
	s = input();
	memcpy(drv->fn, s, 80);
}
static void
set_image_file(Image *img)
{
	char *s;
	printf("New image file (without .dsk) : ");
	s = input();
	strcpy(img->fname, s);
	if (!strchr(img->fname, '.'))
		strcat(img->fname, ".dsk");
	printf("Image file set to %s\n", img->fname);
}
static void
format_image(Drive *drv, Image *img, Disk_info *di, Track_info *ti)
{
	printf("formatting image file %s\n", img->fname);
	if ((img->f = fopen(img->fname, "r"))) {
		fclose(img->f);
		printf("file already exist. Overwrite (y/n) ? ");
		{
			char *s = input();
			if (upcase(s[0]) != 'Y')
				return;
		}
	}
	if (!(img->f = fopen(img->fname, "w"))) {
		printf("error opening %s\n", img->fname);
		return;
	}
	select_format_form(di, ti);

	if (format_to_img(drv, di, ti, img))
		printf("Operation terminated incorrectly.\n");
	else
		printf("formatting ok.\n");
	fclose(img->f);
}
static void
copy_drive_image(Drive *drv, Image *img, Disk_info *di, Track_info *ti)
{
	printf("Copy 'dd' file -> Image\n");
	printf("reading from 'dd' file %s.\n", drv->fn);
	printf("writing to image file %s\n", img->fname);
	if (!(drv->fp=fopen(drv->fn, "r"))) {
		fprintf(stderr, "error: can't open %s\n", drv->fn);
		exit(1);
	}
	if ((img->f = fopen(img->fname, "r"))) {
		fclose(img->f);
		printf("file already exist. Overwrite (y/n) ? ");
		{
			char *s = input();
			if (upcase(s[0]) != 'Y')
				return;
		}
	}
	if (!(img->f = fopen(img->fname, "w"))) {
		printf("error opening %s\n", img->fname);
		return;
	}
	select_format(di, ti);

	if (drv_to_img(drv, di, ti, img))
		printf("Operation terminated incorrectly.\n");
	else
		printf("transfer ok.\n");
	fclose(img->f);
	fclose(drv->fp);
}
static void
copy_image_drive(Drive *drv, Image *img, Disk_info *di, Track_info *ti)
{
	printf("Copy Image -> 'dd' file\n");
	printf("reading from image file %s\n", img->fname);
	if (!(img->f = fopen(img->fname, "r"))) {
		printf("error opening %s\n", img->fname);
		return;
	}
	if (!(drv->fp=fopen(drv->fn, "w"))) {
		fprintf(stderr, "error: can't open %s\n", drv->fn);
		exit(1);
	}
	printf("writing to 'dd' file %s ...\n", drv->fn);

	if (img_to_drv(drv, di, ti, img))
		printf("Operation terminated incorrectly.\n");
	else 
	printf("transfer ok.\n");
	fclose(img->f);
	fclose(drv->fp);
}
static Disk_info *di;
static Track_info *ti;
static Drive *drv;
static Image *img;
static void
menu(void)
{
	char *s;
	printf("\n\n");
	printf("DDTRANS v0.1\n");
	printf("(Transfer CPC/+3/PCW-disks to disk-images and vice versa)\n");
	printf("(and make newly formatted disks)");
	do {
		printf("\n\n");
		printf("1) Set 'dd' file\n");
		printf("2) Set Image file\n");
		printf("3) Copy 'dd' file -> Image\n");
		printf("4) Copy Image -> 'dd' file\n");
		printf("5) Format disk\n");
		printf("6) Quit\n\n");
		printf("Your choice: ");
		s = input();
		switch (s[0]) {
			case '1' :
				set_drive(drv);
				break;
			case '2' :
				set_image_file(img);
				break;
			case '3' :
				copy_drive_image(drv, img, di, ti);
				break;
			case '4' :
				copy_image_drive(drv, img, di, ti);
				break;
			case '5' :
				format_image(drv, img, di, ti);
				break;
			case '6' :
				break;
			default :
				printf(" only 1-6 : ");
		}
	} while (s[0] != '6');
}
int
main()
{
	static char defaultdisc[] = "image.p3d\n";
	if (!(drv = (Drive *) malloc(sizeof(Drive)))) {
		fprintf(stderr, "malloc: no memory\n");
		exit(1);
	}
	if (!(drv->fn = malloc(80))) {
		fprintf(stderr, "malloc: no memory\n");
		exit(1);
	}
	memcpy(drv->fn, defaultdisc, 9);
	if (!(img = (Image *)malloc(sizeof(Image)))) {
		fprintf(stderr, "malloc: no memory\n");
		exit(1);
	}
	strcpy(img->fname, "default.dsk");
	if (!(di = (Disk_info *)malloc(sizeof(Disk_info)))) {
		fprintf(stderr, "malloc: no memory\n");
		exit(1);
	}
	if (!(ti = (Track_info *)malloc(sizeof(Track_info)))) {
		fprintf(stderr, "malloc: no memory\n");
		exit(1);
	}
	menu();
	return 0;
}

