/********************************************************************************
 * Copyright (c) Razvan Surdulescu 1996
 *               Erik Kunze        1996, 1997
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Razvan Surdulescu
 *         Erik Kunze
 *
 * changed by EKU
 *******************************************************************************/
#ifndef lint
static char rcsid[] = "$Id: dialog.c,v 3.23 1997/04/15 04:19:05 erik Rel $";
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include "z80.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "io.h"
#include "util.h"
#ifdef AUDIO
#include "audio.h"
#endif
#ifdef XZX_PLUS3
#ifdef REGISTERED
#include "fdi.h"
#else
#include "fdc.h"
#endif
#endif
#include "emul.h"
#ifdef XZX_IF1
#include "if1.h"
#endif
#ifdef JOY
#include "joystick.h"
#endif
#include "loadsave.h"
#include "monitor.h"
#include "screen.h"
#include "snapshot.h"
#include "tables.h"
#include "version.h"
#include "dialog.h"

#define PAGE_SIZE			14
#define PARAMETER_COLUMN	24
#define MAXFILELEN			256
#define DIR_TYPE			0
#define FILE_TYPE			1
typedef struct _mydirent {
	int type;
	char * name;
} mydirent;

static uns16 xyToAddr (uns8, uns8);
static uns16 xyToAttr (uns8, uns8);
static void saveScreen(void);
static void restoreScreen(void);
static int direntsCompare(void *, void *);
static void direntDestroy(void *);
static int readDirectory(char *, char *, char *);
static mydirent *getFileOrDir(int);
static void displayFiles(int);
static char *addExtension(char *, char *);
static void displayMenuTitle(char *);
static int displayKeyboardLayout(void);
static void genericMenu(void);
#ifdef AUDIO
static void soundMenu(void);
#endif
static void joyMenu(void);
#if defined(XZX_IF1) || defined(XZX_PLUS3)
static void diskMenu(void);
#endif
static void tapeMenu(void);
static char *getBaseName(char *);

int InDialog = 0;
KeySym KeyDialog;


static char *menuBottom[] = {
	"   UP/DOWN arrows move cursor   ",
	"   ENT/KEY select               ",
	"   ESC     return               ",
	NULL
};
static char *mainMenuTxt[] = {
	"Main Menu",
	"Keyboard layout             F1",
	"Save snapshot               F2",
	"Load snapshot               F3",
	"Save screenshot             F4",
	"Tape Options",
	"Reset                       F5",
	"NMI                         F6",
	"Debugger                    F7",
	"Generic Options",
#ifdef AUDIO
	"Sound Options",
#else
	"`Sound Options",
#endif
	"Joystick Options",
#if defined(XZX_IF1) || defined(XZX_PLUS3)
	"Interface I/+3 Options",
#else
	"`Interface I/+3 Options",
#endif
	"Choose Architecture         F9",
	"Quit                       F10",
	NULL
};
static char *architMenuTxt[] = {
	"Architecture Options",
	"Spectrum 48K  (reset)",
	"Spectrum 128k (reset)",
#ifdef XZX_PLUS3
	"Spectrum +3   (reset)",
#else
	"`Spectrum +3   (reset)",
#endif
	"",
	"Spectrum 48K  (no reset)",
	"Spectrum 128k (no reset)",
#ifdef XZX_PLUS3
	"Spectrum +3   (no reset)",
#else
	"`Spectrum +3   (no reset)",
#endif
	"",
#ifdef XZX_IF1
	"Interface I           [ ]",
#else
	"`Interface I           [ ]",
#endif
#ifdef XZX_MF128
	"Multiface 128         [ ]",
#else
	"`Multiface 128         [ ]",
#endif
	NULL
	};
static char *genericMenuTxt[] = {
	"Generic Options",
	"Flashing              [ ]",
	"Border                [ ]",
	"Tstates/frame         [      ]",
	"Ms/frame              [      ]",
	"Fast                  [ ]",
#ifdef SLOWDOWN
	"Slowdown              [      ]",
#else
	"`Slowdown              [      ]",
#endif
	"Refresh               [      ]",
	"Issue                 [ ]",
#ifdef DEBUG
	"Debug                 [      ]",
#else
	"`Debug                 [      ]",
#endif
	NULL
};
#ifdef AUDIO
static char *soundMenuTxt[] = {
	"Sound Options",
#if defined(PCSPKR_AUDIO) || defined(SUN_AUDIO)
	"Speaker               [ ]",
#else
	"`Speaker               [ ]",
#endif
#ifdef AYCHIP_AUDIO
	"AY8912                [ ]",
#else
	"`AY8912                [ ]",
#endif
#if defined(SUN_AUDIO) || defined(XBELL_AUDIO)
	"Volume                [    ]%",
#else
	"`Volume                [    ]%",
#endif
	NULL
};
#endif
static char *joyMenuTxt[] = {
	"Joystick Options",
	"Up                    [      ]",
	"Down                  [      ]",
	"Left                  [      ]",
	"Right                 [      ]",
	"Fire                  [      ]",
	"",
#ifdef JOY
	"Tolerance             [      ]",
	"Calibrate",
#else
	"`Tolerance             [      ]",
	"`Calibrate",
#endif
	NULL
};
#if defined(XZX_IF1) || defined(XZX_PLUS3)
static char *diskMenuTxt[] = {
	"Interface I/+3 Options",
#ifdef XZX_IF1
	"Select microdrive     [ ]",
	"Insert cartridge",
	"Take cartridge out",
	"",
	"Translate NL          [ ]",
	"Strip CR              [ ]",
#else
	"`Select microdrive     [ ]",
	"`Insert cartridge",
	"`Take cartridge out",
	"",
	"`Translate NL          [ ]",
	"`Strip CR              [ ]",
#endif
	"",
#ifdef XZX_PLUS3
	"Select drive          [ ]",
	"Insert disk",
	"Take disk out",
#else
	"`Select drive          [ ]",
	"`Insert disk",
	"`Take disk out",
#endif
	NULL
};
#endif
static char *tapeMenuTxt[] = {
	"Tape Options",
	"Select inputfile",
	"Select outputfile",
	"Browse through inputfile",
	NULL
};
static uns8 cursorX = 0;
static uns8 cursorY = 0;
static uns8 screenAttr = MENU_TEXT_ATTR;

static uns8 *screenBackup = NULL;
static int saveScreenSelect;
static char *saveDir;
static char *saveFSelDir;
#ifdef AYCHIP_AUDIO
static int oldAy;
#endif
static int numFiles = 0;
static int numDirectories = 0;
static llist *dirDirectoriesList = NULL;

static llist *dirFilesList = NULL;

void
DialogInit(int copyright)
{
	char *ver;
	time_t timeStart;
	if (!(saveDir = getcwd(NULL, MAXPATHLEN)))
	{
		Msg(M_FATAL, "getcwd failed");
	}
	screenBackup = (uns8 *)Malloc(PIXEL_LENGTH * ATTR_LENGTH, "DialogInit");
	saveFSelDir = (char*)Malloc(MAXPATHLEN, "DialogInit");
	(void)strcpy(saveFSelDir, saveDir);
	if (copyright)
	{

		ClearScreen();
		SetBorderColor((MENU_TEXT_ATTR & PAPER) >> 3);
		ver = (char *)Malloc(sizeof("XZX X.XX.XX") + 1, "DialogInit");
		(void)sprintf(ver, "XZX %d.%d.%d", ZX_Major, ZX_Minor, ZX_Patch);
		SetCursor((COLS - strlen(ver)) / 2, 5);
		PrintString(ver);
		free(ver);
		SetCursor(1, 7);
		PrintString("ZX Spectrum 48/128/+3 emulator");
		SetCursor(1, 11);
		PrintString(" \177 Des Herriott       1993-94");
		SetCursor(1, 13);
		PrintString("   Erik Kunze         1995-97");
		SetCursor(1, 15);
		PrintString("   Razvan Surdulescu  1996   ");

		InDialog = 1;
		timeStart = time(NULL);
		while((time(NULL) - timeStart) < 5)
		{
			ScreenRefresh();
			CheckEvents();
		}
		InDialog = 0;
		SetBorderColor(LastBorderColor);
	}
}

static uns16
xyToAddr (uns8 x, uns8 y)
{
	assert(x < COLS &&  y < ROWS);
	y *= 8;
	return (DFILE +
			(y / 64) * (PIXEL_LENGTH / 3) +
			((y % 64) / 8) * COLS +
			(y % 8) * X_PIXELS +
			x);
}

static uns16
xyToAttr (uns8 x, uns8 y)
{
	assert(x < COLS && y < ROWS);
	return (DFILE + ATTR_OFFSET + y * COLS + x);
}

void
SetAttr (uns8 attr)
{
	screenAttr = attr;
}

uns8
GetAttr(void)
{
    return screenAttr;
}

void
SetCursor (uns8 x, uns8 y)
{
	cursorX = x < COLS ? x : COLS - 1; 
	cursorY = y < ROWS ? y : ROWS - 1;
}

void
GetCursor(uns8 *x, uns8 *y)
{
	*x = cursorX;
	*y = cursorY;
}

void 
ClearScreen(void)
{
	cursorX = 0;
	cursorY = 0;
	DrawBox(COLS, ROWS);
}

void 
DrawBox(uns8 width, uns8 height)
{
	uns8 saveX = cursorX, saveY = cursorY;
	uns8 i;
	for (; height != 0; height--, cursorY++)
	{
		for (cursorX = saveX, i = width; i != 0; i--)
		{
			PrintLetter(' ');
		}
	}
	cursorX = saveX;
	cursorY = saveY;
}

void 
PrintLetter(uns8 letter)
{
	if (cursorX < COLS && cursorY < ROWS)
	{
		uns8 i;
		uns16 addr = xyToAddr(cursorX, cursorY);
		uns8 *chr;
		if (letter < ' ' || letter > ' ' + sizeof(CharSet)/sizeof(CharSet[0]))
		{
			letter = '?';
		}
		chr = (uns8 *)CharSet[letter - ' '];
		WR_BYTE(xyToAttr(cursorX, cursorY), screenAttr);
		for (i = 0; i < 8; i++, addr += X_PIXELS, chr++)
		{
			WR_BYTE(addr, *chr);
		}
		cursorX++;
	}
}

void 
PrintString(char *string)
{
	while (*string && cursorX < COLS)
	{
		PrintLetter(*string++);
	}
}

KeySym
GetKey(void)
{
	for (KeyDialog = 0; !KeyDialog;)
	{
		CheckEvents();
	}
	return KeyDialog;
}

int
GetString(char *string, int len, int width)
{
	int oldInDialog = InDialog;
	int finish = 0;
	int pos;
	uns8 begX = cursorX;
	uns8 begY = cursorY;
	InDialog = 1;

	PrintLetter('_');
	for (pos = width; pos > 1; pos--)
	{
		PrintLetter(' ');
	}
	pos = 0;
	SetCursor(begX, begY);
	ScreenRefresh();
	for (;;)
	{
		switch (GetKey())
		{
			case XK_Return:
				finish = 1;
				goto quit;
				break;
			case ESCAPE:
			case XK_Tab:
				finish = 2;
				goto quit;
				break;
			case XK_BackSpace:
				if (pos > 0)
				{
					string[--pos] = '\0';
					if (pos < width - 1) 
					{
						PrintLetter(' ');
						cursorX -= 2;
					}
					else
					{
						SetCursor(begX, begY);
						PrintString(string + (pos - width) + 1);
					}
				}
				break;
			default:
				if (KeyDialog >= 0x20 && KeyDialog <= 0x7e && pos + 1 < len)
				{
					string[pos++] = (char)KeyDialog;
					string[pos] = '\0';
					if (pos < width)
					{
						PrintLetter(KeyDialog);
					}
					else
					{
						SetCursor(begX, begY);
						PrintString(string + (pos - width) + 1);
					}
				}
				break;
		}
		PrintLetter('_');
		cursorX--;
		ScreenRefresh();
	}
  quit:

	PrintLetter(' ');
	string[pos] = '\0';
	InDialog = oldInDialog;
	return (finish != 2);
}

int
ConfirmBox(char *string)
{

	int oldInDialog = EnterOSD();
	KeySym key;
	ClearScreen();
	SetCursor((COLS - strlen(string)) / 2, ROWS / 2 - 1);
	PrintString(string);
	ScreenRefresh();
	key = GetKey();

	LeaveOSD(oldInDialog);
	return (key == XK_y || key == XK_Y);
}

int
EnterOSD(void)
{
	int oldInDialog = InDialog;

	if (!oldInDialog)
	{
		XAutoRepeatOn(Dpy);
		saveScreenSelect = ScreenSelect;
		ScreenSelect = RAM5;
		saveScreen();
#ifdef AYCHIP_AUDIO
		if ((oldAy = AyOn))
		{
			AyOnOff();
		}
#endif
	}
	InDialog = 1;
	KeyDialog = 0;
	return oldInDialog;
}

void
LeaveOSD(int oldInDialog)
{

	if (!oldInDialog)
	{
		restoreScreen();
		ScreenSelect = saveScreenSelect;
		ForceScreenRefresh();
#ifdef AYCHIP_AUDIO
		if (oldAy)
		{
			AyOnOff();
		}
#endif
	}
	InDialog = oldInDialog;
}

static void
saveScreen(void)
{
	(void)memcpy(screenBackup, RealMemory[ScreenSelect],
				 PIXEL_LENGTH + ATTR_LENGTH);
	SetBorderColor((MENU_TEXT_ATTR & PAPER) >> 3);
}

static void
restoreScreen(void)
{
	(void)memcpy(RealMemory[ScreenSelect], screenBackup,
				 PIXEL_LENGTH + ATTR_LENGTH);
	SetBorderColor(LastBorderColor);
}

static int
direntsCompare(void *d1, void *d2)
{
	return (strcmp(((mydirent *)d1)->name, ((mydirent *)d2)->name));
}

static void
direntDestroy(void *d)
{
	assert(((mydirent *)d)->name);
	free(((mydirent *)d)->name);
}

static int
readDirectory(char *fext1, char *fext2, char * fext3)
{
	DIR *dir;
	struct dirent *ent;
	struct stat buf;
	char *currentPath, *p;
	mydirent *myDirent;
	if (!(currentPath = getcwd(NULL, MAXPATHLEN))
		|| !(dir = opendir(currentPath)))
	{
		Msg(M_PERR, "couldn't open directory <%s>", currentPath);
		free(currentPath);
		return -1;
	}
	while ((ent = readdir(dir)))
	{

		if (ent->d_name[0] == '.' && ent->d_name[1] != '.')
		{
			continue;
		}
		(void)stat(ent->d_name, &buf);
		if (S_ISDIR(buf.st_mode))
		{

			myDirent = Malloc(sizeof(mydirent), "readDirectory");
			myDirent->type = DIR_TYPE;
			myDirent->name = Strdup(ent->d_name, "readDirectory");
			dirDirectoriesList = SortedInsertList((void *)myDirent, 
												  direntsCompare, 
												  dirDirectoriesList);
			numDirectories++;
		}
		else if (S_ISREG(buf.st_mode))
		{

			if (!fext1
				|| ((p = strrchr(ent->d_name, '.')) != NULL
					&& (!strcasecmp(++p, fext1)
						|| (fext2 && !strcasecmp(p, fext2))
						|| (fext3 && !strcasecmp(p, fext3)))))
			{

				myDirent = Malloc(sizeof(mydirent), "readDirectory");
				myDirent->type = FILE_TYPE;
				myDirent->name = Strdup(ent->d_name, "readDirectory");
				dirFilesList = SortedInsertList((void *)myDirent,
												direntsCompare,
												dirFilesList);
				numFiles++;
			}
		}
	}
	(void)closedir(dir);
#ifdef DEBUG
	if (GETCFG(debug) & D_OSD)
	{
		Msg(M_DEBUG, "%s: %d dirs, %d files", currentPath, numDirectories,
			numFiles);
	}
#endif
	free(currentPath);
	return 0;
}

static mydirent *
getFileOrDir(int which) 
{
	assert(which >= 0);
	if (which < numDirectories)
	{
		return ((mydirent*) RetrieveElemList(dirDirectoriesList, which));
	}
	which -= numDirectories;
	if (which < numFiles)
	{
		return ((mydirent*) RetrieveElemList(dirFilesList, which));
	}
	return NULL;
}

static void
displayFiles(int page)
{
	int cnt, len;
	char *name;
	mydirent *disp;
	uns8 attr= GetAttr();
	SetAttr(MENU_SELECT_ATTR);
	SetCursor(MENU_CURSOR_X + 1, MENU_CURSOR_Y);
	DrawBox(COLS - 2 * (MENU_CURSOR_X + 1), PAGE_SIZE);

	for (cnt = page * PAGE_SIZE; cnt < (page + 1) * PAGE_SIZE; cnt++)
	{
		if (!(disp = getFileOrDir(cnt)))
		{
			break;
		}
		name  = disp->name;
		for (len = COLS - 2 * (MENU_CURSOR_X + 1); len > 0 && *name; len--)
		{
			PrintLetter(*name++);
		}
		if (disp->type == DIR_TYPE && len)
		{
			PrintLetter('/');
		}
		SetCursor(MENU_CURSOR_X + 1, cursorY + 1);
	}

	SetAttr(attr);
}

static char *
addExtension(char *string, char *ext)
{
    char *p;

	if (!(p = strrchr(string, '.')))
	{
		p = (char*)Malloc(strlen(string) + strlen(ext) + 2, "addExtension");
		(void)strcpy(p, string);
		(void)strcat(p, ".");
		(void)strcat(p, ext);
	}
	else
	{
		p = Strdup(string, "addExtension");
	}
	return p;
}

char *
FileSelector(char *header, int valid, char * fext1, char * fext2, char * fext3)
{
	static int item = 0;
	int oldInDialog;
	mydirent *file;
	struct stat buf;
	char name[MAXFILELEN];
	char *returnName = NULL;
	char *returnTotal = NULL;

	oldInDialog = EnterOSD();

	(void)chdir(saveFSelDir);

	{
		char *selectorText[PAGE_SIZE + 2];
		int i;
		selectorText[0] = header;
		for (i = 1; i <= NOPTS(selectorText); i++)
		{
			selectorText[i] = " ";
		}
		selectorText[i] = NULL;
		DisplayMenu(selectorText);
	}
	SetCursor(0, ROWS - NOPTS(menuBottom) - 2);
	PrintString("   TAB     switch fields        ");
	ScreenRefresh();

	if (readDirectory(fext1, fext2, fext3) == -1)
	{
		goto quit;
	}

	if (item >= numFiles + numDirectories)
	{
		item = numFiles + numDirectories - 1;
	}
	displayFiles(item / PAGE_SIZE);
	SetCursor(0, PAGE_SIZE + 3);
	PrintString(" File [                       ] ");
	for (;;)
	{
		switch (BrowseList(&item, numFiles + numDirectories, PAGE_SIZE,
						   displayFiles))
		{

			case XK_Tab:

				{
					int ret;
					SetAttr(MENU_SELECT_ATTR);
					SetCursor(7, PAGE_SIZE + 3);
					ret = GetString(name, MAXFILELEN, 23);
					SetAttr(MENU_TEXT_ATTR);
					SetCursor(0, PAGE_SIZE + 3);
					PrintString(" File [                       ] ");
					if (ret)
					{
						if (stat(name, &buf) != -1)
						{
							if (S_ISDIR(buf.st_mode))
							{
								(void)chdir(name);
								goto chdir;
							}
							else if (S_ISREG(buf.st_mode))
							{
								if (valid
									|| ConfirmBox("Overwrite file ? [y/n]"))
								{
									returnName = addExtension(name, fext1);
								}
								goto quit;
							}
						}
						else if (errno == ENOENT && !valid)
						{
							returnName = addExtension(name, fext1);
							goto quit;
						}
					}
				}
				break;
			case XK_Return:

				file = getFileOrDir(item);
				assert(file);
				if (file->type == DIR_TYPE)
				{
					(void)chdir(file->name);
				  chdir:
					DestroyList(dirFilesList, direntDestroy);
					numFiles = 0;
					dirFilesList = NULL;
					DestroyList(dirDirectoriesList, direntDestroy);
					numDirectories = 0;
					dirDirectoriesList = NULL;
					if (readDirectory(fext1, fext2, fext3) == -1)
					{
						goto quit;
					}
					displayFiles(0);
					item = 0;
				}
				else
				{
					assert(file->type == FILE_TYPE);
					if (valid || ConfirmBox("Overwrite file ? [y/n]"))
					{
						returnName = Strdup(file->name, "FileSelector");
					}
					goto quit;
				}
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
	DestroyList(dirFilesList, direntDestroy);
	numFiles = 0;
	dirFilesList = NULL;
	DestroyList(dirDirectoriesList, direntDestroy);
	numDirectories = 0;
	dirDirectoriesList = NULL;

	if (!getcwd(saveFSelDir, MAXPATHLEN))
	{
		Msg(M_FATAL, "getcwd failed");
	}

	if (returnName)
	{
		if (*returnName == '/'
			|| !strncmp(returnName, "./", 2)
			|| !strncmp(returnName, "../", 3))
		{

			returnTotal = Strdup(returnName, "FileSelector");
		}
		else
		{

			returnTotal = (char*)Malloc(strlen(saveFSelDir) +
										strlen(returnName) + 2, "FileSelector");
			(void)strcpy(returnTotal, saveFSelDir);
			(void)strcat(returnTotal, "/");
			(void)strcat(returnTotal, returnName);
		}
		free(returnName);
	}

	(void)chdir(saveDir);
	return returnTotal;
}

static void
displayMenuTitle(char *title)
{
	uns8 attr= GetAttr();

	SetAttr(BRIGHT | (BLACK << 3) | WHITE);
	PrintLetter(' ');
	PrintString(title);

	while (cursorX < COLS)
	{
		PrintLetter(' ');
	}

	if (!GETCFG(mono))
	{
		SetCursor(COLS - 6, cursorY);
		SetAttr(BRIGHT | (BLACK << 3) | RED);
		PrintLetter('\200');
		SetAttr(BRIGHT | (RED << 3) | YELLOW);
		PrintLetter('\200');
		SetAttr(BRIGHT | (YELLOW << 3) | GREEN);
		PrintLetter('\200');
		SetAttr(BRIGHT | (GREEN << 3) | CYAN);
		PrintLetter('\200');
		SetAttr(BRIGHT | (CYAN << 3) | BLACK);
		PrintLetter('\200');
	}

	SetAttr(attr);
}

void
DisplayMenu(char **menu)
{
	int i;
	char **s;
	SetAttr(MENU_TEXT_ATTR);
	ClearScreen();

	displayMenuTitle(*menu++);

	SetAttr(MENU_SELECT_ATTR);
	while (*menu)
	{

		SetCursor(0, cursorY + 1);
		PrintLetter('\201');

	    if (*menu[0] == '`')
		{
			SetAttr(MENU_DISABLED_ATTR);
			PrintString(&menu[0][1]);
			SetAttr(MENU_SELECT_ATTR);
		}
		else
		{
			PrintString(menu[0]);
		}
		while (cursorX < COLS - 1)
		{
			PrintLetter(' ');
		}

		PrintLetter('\205');
		menu++;
	}

	SetCursor(0, cursorY + 1);
	PrintLetter('\202');
	for (i = COLS - 2; i > 0; i--)
	{
		PrintLetter('\203');
	}
	PrintLetter('\204');

	SetAttr(MENU_TEXT_ATTR);
	SetCursor(0, ROWS - NOPTS(menuBottom) - 1);
	for (s = menuBottom; *s; s++)
	{
		PrintString(*s);
		SetCursor(0, cursorY + 1);
	}
	ScreenRefresh();
}

KeySym
BrowseList(int *curItem, int maxItems, int itemsPerPage, void (*printPage)(int))
{
	int i;
	uns16 mem;
	int item, page;
	KeySym key;
	assert(*curItem >= 0 && *curItem < maxItems);
	item = *curItem % itemsPerPage;
	page = *curItem / itemsPerPage;
	for (;;)
	{

		SetCursor(MENU_CURSOR_X, MENU_CURSOR_Y + item);
		mem = xyToAttr(cursorX, cursorY);
		if (GETCFG(mono))
		{
			for (i = COLS; i > 0; i--, mem++)
			{
				WR_BYTE(mem, Reverse[RD_BYTE(mem) & ~FLASH]);
			}
		}
		else
		{
			for (i = COLS; i > 0; i--, mem++)
			{
				WR_BYTE(mem,(RD_BYTE(mem) & ~PAPER) | (CYAN << 3));
			}
		}
		ScreenRefresh();
		key = GetKey();

		mem = xyToAttr(MENU_CURSOR_X, MENU_CURSOR_Y + item);
		if (GETCFG(mono))
		{
			for (i = COLS; i > 0; i--, mem++)
			{
				WR_BYTE(mem, Reverse[RD_BYTE(mem) & ~FLASH]);
			}
		}
		else
		{
			for (i = COLS; i > 0; i--, mem++)
			{
				WR_BYTE(mem,(RD_BYTE(mem) & ~PAPER) | (WHITE << 3) | BRIGHT);
			}
		}
		ScreenRefresh();
		switch (key)
		{
			case XK_Prior:
				if (--page < 0)
				{
					goto home;
				}
				if (printPage)
				{
					printPage(page);
				}
				break;
			case XK_Next:

				if (item + ++page * itemsPerPage >= maxItems)
				{
					goto end;
				}
				if (printPage)
				{
					printPage(page);
				}
				break;
			case XK_Up:
				if (--item < 0)
				{
					if (page > 0)
					{

						item = itemsPerPage - 1;
						page--;
						if (printPage)
						{
							printPage(page);
						}
					}
					else
					{
						goto end;
					}
				}
				break;
			case XK_Home:
		  home:
				item = 0;
				page = 0;
				if (printPage)
				{
					printPage(page);
				}
				break;
			case XK_Down:

				if (++item + page * itemsPerPage >= maxItems)
				{
					goto home;
				}
				if (item > itemsPerPage - 1)
				{
					if (page < maxItems / PAGE_SIZE)
					{

						item = 0;
						page++;
						if (printPage)
						{
							printPage(page);
						}
					}
					else
					{
						goto home;
					}
				}
				break;
			case XK_End:
			  end:
				item = (maxItems - 1) % itemsPerPage;
				page = (maxItems - 1) / itemsPerPage;
				if (printPage)
				{
					printPage(page);
				}
				break;
			default:
				goto quit;
		}
	}
  quit:
	*curItem = item + page * itemsPerPage;
	return KeyDialog;
}

static int
displayKeyboardLayout()
{
	FILE *ifp;
	size_t nread;
	if (!(ifp = OpenXzxFile(GETCFG(kbdlayout), "r")))
	{
		Msg(M_PERR, "couldn't open keyboard layout image <%s> for reading",
			GETCFG(kbdlayout));
		return -1;
	}
	nread = fread(RealMemory[ScreenSelect], 1, PIXEL_LENGTH + ATTR_LENGTH, ifp);

	if (nread != PIXEL_LENGTH + ATTR_LENGTH)
	{
		Msg(M_WARN, "layout image is only %d bytes", nread);
	}
	(void)fclose(ifp);
	SetBorderColor(BLACK);
	ForceScreenRefresh();
	ScreenRefresh();
	return 0;
}

void 
MainMenu(int choice)
{
	int item = choice;
	int oldInDialog;
	char *fname;
	assert(choice >= 0 && choice < NOPTS(mainMenuTxt));

	oldInDialog = EnterOSD();

	DisplayMenu(mainMenuTxt);
	for (;;)
	{
		switch (BrowseList(&item, NOPTS(mainMenuTxt), NOPTS(mainMenuTxt), NULL))
		{

			case XK_F1:
				item = 0;
				break;
			case XK_F2:
				item = 1;
				break;
			case XK_F3:
				item = 2;
				break;
			case XK_F4:
				item = 3;
				break;
			case XK_F5:
				item = 5;
				break;
			case XK_F6:
				item = 6;
				break;
			case XK_F7:
				item = 7;
				break;
			case XK_F9:
				item = 12;
				break;
			case XK_F10:
			case XK_F12:
				item = 13;
				break;
			case XK_Return:

				switch(item)
				{
					case 0:
						if (displayKeyboardLayout() != -1)
						{

							(void)GetKey();
							SetBorderColor((MENU_TEXT_ATTR & PAPER) >> 3);
						}
						break;
					case 1:
						if((fname = FileSelector("Save Snapshot", False, "z80",
												 GETCFG(machine) == 48 ?
												 "sna" : InSlt ?
												 "slt" : NULL,
												 GETCFG(machine) == 48 &&
												 InSlt ? "slt" : NULL)))
						{
							restoreScreen();
							(void)WriteSnapshot(fname);
							free(fname);
							saveScreen();
						}
						break;
					case 2:
						if ((fname = FileSelector("Load Snapshot", True, "z80",
												  "sna", "slt")))
						{
							(void)ReadSnapshot(fname);
							free(fname);
							saveScreen();
						}
						break;
					case 3:
						if((fname = FileSelector("Save Screenshot", False, "scr",
												 NULL, NULL)))
						{
							restoreScreen();
							(void)WriteScreenDump(fname);	  
							free(fname);
						}
						break;
					case 4:
						tapeMenu();
						break;
					case 5:
						Z80_Reset();
						break;
					case 6:
						Z80_NMI();
						break;
					case 7:
						Debugger(DBG_OSD);
						break;
					case 8:
						genericMenu();
						break;
#ifdef AUDIO
					case 9:
						soundMenu();
						break;
#endif
					case 10:
						joyMenu();
						break;
#if defined(XZX_IF1) || defined(XZX_PLUS3)
					case 11:
						diskMenu();
						break;
#endif
					case 12:
						HardwareMenu();
						break;
					case 13:
						Quit(0);
						break;
				}
				if (item == choice)
				{
					goto quit;
				}
				DisplayMenu(mainMenuTxt);
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
}

static void 
genericMenu(void)
{
	int item = 0;
	int oldInDialog;
	char buffer[7];
	int val;

	oldInDialog = EnterOSD();

	DisplayMenu(genericMenuTxt);
	for (;;)
	{
		SetAttr(MENU_SELECT_ATTR);
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y);
		PrintLetter(GETCFG(flashing) ? 'Y' : 'N');
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 1);
		PrintLetter(GETCFG(border) ? 'Y' : 'N');
		(void)sprintf(buffer, "%-6d", GETCFG(cycles));
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 2);
		PrintString(buffer);
		(void)sprintf(buffer, "%-6d", GETCFG(ms));
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 3);
		PrintString(buffer);
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 4);
		PrintLetter(GETCFG(fast) ? 'Y' : 'N');
#ifdef SLOWDOWN
		(void)sprintf(buffer, "%-6d", GETCFG(slowdown));
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 5);
		PrintString(buffer);
#endif
#ifdef MITSHM
		if (GETCFG(mitshm))
		{
			(void)sprintf(buffer, "%-6d", GETCFG(rrShm));
		}
		else
#endif
		{
			(void)sprintf(buffer, "%-6d", GETCFG(rrNoShm));
		}
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 6);
		PrintString(buffer);
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 7);
		PrintLetter('0' + GETCFG(issue));
#ifdef DEBUG
		(void)sprintf(buffer, "%-6d", GETCFG(debug));
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 8);
		PrintString(buffer);
#endif
		SetAttr(MENU_TEXT_ATTR);
		switch (BrowseList(&item, NOPTS(genericMenuTxt), NOPTS(genericMenuTxt),
						   NULL))
		{

			case XK_Return:

				switch(item)
				{
					case 0:
						SETCFG(flashing, !GETCFG(flashing));   
						break;
					case 1:
						SETCFG(border, !GETCFG(border));   
						break;
					case 2:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						if (GetString(buffer, 6, 6))
						{
							val = atoi(buffer);
							if (val >= 1000)
							{
								SETCFG(cycles, val);
								CyclesPerLine = GETCFG(cycles) / TVLINES;
#ifdef SUN_AUDIO
								SpeakerReInit();
#endif
							}
						}
						SetAttr(MENU_TEXT_ATTR);
						break;
					case 3:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						if (GetString(buffer, 6, 6))
						{
							val = atoi(buffer);
							if (val > 0 && val <= 100)
							{
								SETCFG(ms, val);
								IntFrequency(1000 * GETCFG(ms));
#ifdef SUN_AUDIO
								SpeakerReInit();
#endif
							}
						}
						SetAttr(MENU_TEXT_ATTR);
						break;
					case 4:
						SETCFG(fast, !GETCFG(fast));
						break;
#ifdef SLOWDOWN
					case 5:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						if (GetString(buffer, 6, 6))
						{
							val = atoi(buffer);
							if (val >= 0 && val <= 9999)
							{
								SETCFG(slowdown, val);
							}
						}
						SetAttr(MENU_TEXT_ATTR);
						break;
#endif
					case 6:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						if (GetString(buffer, 6, 6))
						{
							val = atoi(buffer);
							if (val > 0 && val <= 50)
							{
#ifdef MITSHM
								if (GETCFG(mitshm))
								{
									SETCFG(rrShm, val);
								}
								else
#endif
								{
									SETCFG(rrNoShm, val);
								}
							}
						}
						SetAttr(MENU_TEXT_ATTR);
						break;
					case 7:
						SetCursor(PARAMETER_COLUMN, cursorY);
						SETCFG(issue, GETCFG(issue) == 2 ? 3 : 2);
						break;
#ifdef DEBUG
					case 8:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						if (GetString(buffer, 6, 6))
						{
							val = atoi(buffer);
							if (val >= 0 && val <= 99999)
							{
								SETCFG(debug, val);
							}
						}
						SetAttr(MENU_TEXT_ATTR);
						break;
#endif
				}
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
}

void
HardwareMenu(void)
{
	int item = 0;
	int oldInDialog;

	oldInDialog = EnterOSD();

	DisplayMenu(architMenuTxt);
	SetCursor(MENU_CURSOR_X + 1, MENU_CURSOR_Y + 12);
	PrintString("Current mode is ");
	PrintString(GETCFG(machine) == 48 ?
				"48K" : GETCFG(machine) == 128 ? "128K" : "+3");
	for (;;)
	{
		SetAttr(MENU_SELECT_ATTR);
#ifdef XZX_IF1
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 8);
		PrintLetter(GETCFG(machine) == 3 ?
					'-' : GETCFG(if1_active) ? 'Y' : 'N');
#endif
#ifdef XZX_MF128
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 9);
		PrintLetter(GETCFG(machine) == 3 ?
					'-' : GETCFG(mf128_active) ? 'Y' : 'N');
#endif
		SetAttr(MENU_TEXT_ATTR);
		switch (BrowseList(&item, NOPTS(architMenuTxt), NOPTS(architMenuTxt),
						   NULL))
		{

			case XK_Return:

				switch(item)
				{
					case 0:
						(void)SelectModel(48, True);
						goto quit;
						break;
					case 1:
						(void)SelectModel(128, True);
						goto quit;
						break;
#ifdef XZX_PLUS3
					case 2:
						(void)SelectModel(3, True);
						goto quit;
						break;
#endif
					case 4:
						(void)SelectModel(48, False);
						goto quit;
						break;
					case 5:
						(void)SelectModel(128, False);
						goto quit;
						break;
#ifdef XZX_PLUS3
					case 6:
						(void)SelectModel(3, False);
						goto quit;
						break;
#endif
#ifdef XZX_IF1
					case 8:
#ifdef XZX_PLUS3
						if (GETCFG(machine) != 3)
#endif
						{
							If1OnOff(!GETCFG(if1_active));
						}
						break;
#endif
#ifdef XZX_MF128
					case 9:
#ifdef XZX_PLUS3
						if (GETCFG(machine) != 3)
#endif
						{
							SETCFG(mf128_active, !GETCFG(mf128_active));
							if (GETCFG(mf128_active))
							{
								(void)ReadROMs();
							}
						}
						break;
#endif
				}
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
}

#ifdef AUDIO
static void 
soundMenu(void)
{
	int item = 0;
	int oldInDialog;
#if defined(SUN_AUDIO) || defined(XBELL_AUDIO)
	char buffer[4];
	int val;
#endif

	oldInDialog = EnterOSD();

	DisplayMenu(soundMenuTxt);
	for (;;)
	{
		SetAttr(MENU_SELECT_ATTR);
#if defined(PCSPKR_AUDIO) || defined(SUN_AUDIO)
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y);
		PrintLetter(SpOn ? 'Y' : 'N');
#endif
#ifdef AYCHIP_AUDIO
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 1);
		PrintLetter(oldAy ? 'Y' : 'N');
#endif
#if defined(SUN_AUDIO) || defined(XBELL_AUDIO)
		(void)sprintf(buffer, "%-3d", GETCFG(volume));
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 2);
		PrintString(buffer);
#endif
		SetAttr(MENU_SELECT_ATTR);
		switch (BrowseList(&item, NOPTS(soundMenuTxt), NOPTS(soundMenuTxt),
						   NULL))
		{

			case XK_Return:

				switch(item)
				{
#if defined(PCSPKR_AUDIO) || defined(SUN_AUDIO)
					case 0:
#if defined(SUN_AUDIO) && defined(AYCHIP_AUDIO)
						if (oldAy && !SpOn)
						{
							oldAy = 0;
						}
#endif
						SpeakerOnOff();
						break;
#endif
#ifdef AYCHIP_AUDIO
					case 1:
#ifdef SUN_AUDIO
						if (SpOn && !oldAy)
						{
							SpeakerOnOff();
						}
#endif
						oldAy = !oldAy;
						break;
#endif
#if defined(SUN_AUDIO) || defined(XBELL_AUDIO)
					case 2:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						if (GetString(buffer, 4, 4))
						{
							val = atoi(buffer);
							if (val >= 0 && val <= 100)
							{
								SETCFG(volume, val);
							}
						}
						SetAttr(MENU_TEXT_ATTR);
						break;
#endif
				}
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
}
#endif

static void 
joyMenu(void)
{
	int item = 0;
	int oldInDialog;
	char *p;
	int i, j;
#ifdef JOY
	char buffer[7];
	int val;
#endif

	oldInDialog = EnterOSD();

	DisplayMenu(joyMenuTxt);
	for (;;)
	{
		SetAttr(MENU_SELECT_ATTR);
		for (i = 0; i < 5; i++)
		{
			SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + i);
			p = XKeysymToString(GETCFG(joyKeys[i]));
			for (j = 0; j < 6 && *p; j++)
			{
				PrintLetter(*p++);
			}
			for (; j < 6; j++)
			{
				PrintLetter(' ');
			}
		}
#ifdef JOY
		(void)sprintf(buffer, "%-6d", GETCFG(joyTolerance));
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 6);
		PrintString(buffer);
#endif
		SetAttr(MENU_SELECT_ATTR);
		switch (BrowseList(&item, NOPTS(joyMenuTxt), NOPTS(joyMenuTxt), NULL))
		{

			case XK_Return:

				switch(item)
				{
					case 0:
					case 1:
					case 2:
					case 3:
					case 4:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						PrintString("      ");
						SetAttr(MENU_TEXT_ATTR);
						ScreenRefresh();
						SETCFG(joyKeys[cursorY - MENU_CURSOR_Y], GetKey());
						break;
#ifdef JOY
					case 6:
						SetAttr(MENU_SELECT_ATTR);
						SetCursor(PARAMETER_COLUMN, cursorY);
						if (GetString(buffer, 6, 6))
						{
							val = atoi(buffer);
							if (val > 0 && val <= 9999)
							{
								SETCFG(joyTolerance, val);
							}
						}
						SetAttr(MENU_TEXT_ATTR);
						break;
					case 7:
						if (GETCFG(useJoy))
						{
							JoyCalibrate();
							DisplayMenu(joyMenuTxt);
						}
						break;
#endif
				}
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
}

#if defined(XZX_IF1) || defined(XZX_PLUS3)
static void 
diskMenu(void)
{
	int item = 0;
	int oldInDialog;
#ifdef XZX_IF1
	static int mdr = 0;
#endif
#ifdef XZX_PLUS3
	static int disk = 0;
#endif

	oldInDialog = EnterOSD();

	DisplayMenu(diskMenuTxt);
	for (;;)
	{
#ifdef XZX_IF1
		SetAttr(MENU_SELECT_ATTR);
		if (GETCFG(if1_active))
		{
			SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 0);
			PrintLetter((char)('1' + mdr));
			SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 4);
			PrintLetter(GETCFG(translate_nl) ? 'Y' : 'N');
			SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 5);
			PrintLetter(GETCFG(strip_nl) ? 'Y' : 'N');
		}
		else
		{
			SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 0);
			PrintLetter('-');
			SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 4);
			PrintLetter('-');
			SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 5);
			PrintLetter('-');
		}
		SetAttr(MENU_TEXT_ATTR);
		SetCursor(MENU_CURSOR_X + 1, MENU_CURSOR_Y + NOPTS(diskMenuTxt) + 2);
		PrintString("Current cartridge");
		SetCursor(MENU_CURSOR_X + 2, MENU_CURSOR_Y + NOPTS(diskMenuTxt) + 3);
		PrintString(getBaseName(GETCFG(if1_active) ? GetCartName(mdr) : NULL));
		while (cursorX < COLS - 1)
		{
			PrintLetter(' ');
		}
#endif
#ifdef XZX_PLUS3
		SetAttr(MENU_SELECT_ATTR);
		SetCursor(PARAMETER_COLUMN, MENU_CURSOR_Y + 7);
		PrintLetter(GETCFG(machine) == 3 ? (char)('A' + disk) : '-');
		SetAttr(MENU_TEXT_ATTR);
#ifdef XZX_IF1
		SetCursor(MENU_CURSOR_X + 1, MENU_CURSOR_Y + NOPTS(diskMenuTxt) + 5);
		PrintString("Current disk");
		SetCursor(MENU_CURSOR_X + 2, MENU_CURSOR_Y + NOPTS(diskMenuTxt) + 6);
		PrintString(getBaseName(GETCFG(machine) == 3 ?
								GetDiskName(disk) : NULL));
#else
		SetCursor(MENU_CURSOR_X + 1, MENU_CURSOR_Y + NOPTS(diskMenuTxt) + 2);
		PrintString("Current disk");
		SetCursor(MENU_CURSOR_X + 2, MENU_CURSOR_Y + NOPTS(diskMenuTxt) + 3);
		PrintString(getBaseName(GETCFG(machine) == 3 ?
								GetDiskName(disk) : NULL));
#endif
		while (cursorX < COLS - 1)
		{
			PrintLetter(' ');
		}
#endif
	  again:
		switch (BrowseList(&item, NOPTS(diskMenuTxt), NOPTS(diskMenuTxt), NULL))
		{

			case XK_Return:

#ifdef XZX_IF1
				if (GETCFG(if1_active))
				{
					switch(item)
					{
						case 0:
							mdr = ++mdr % MAX_DRIVES;
							break;
						case 1:
							(void)InsertCartridge(mdr);
							DisplayMenu(diskMenuTxt);
							break;
						case 2:
							EjectCartridge(mdr);
							break;
						case 4:
							SETCFG(translate_nl, !GETCFG(translate_nl));
							break;
						case 5:
							SETCFG(strip_nl, !GETCFG(strip_nl));
							break;
						default:
						  goto again;
					}
				}
				else
#endif
#ifdef XZX_PLUS3
				if (GETCFG(machine) == 3)
				{
					switch(item)
					{
						case 7:
							disk = !disk;
							break;
						case 8:
							(void)InsertDisk(disk);
							DisplayMenu(diskMenuTxt);
							break;
						case 9:
							EjectDisk(disk);
							break;
						default:
							goto again;
					}
				}
				else
#endif
				{
					goto again;
				}
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
}
#endif

static void 
tapeMenu(void)
{
	int item = 0;
	int oldInDialog;

	oldInDialog = EnterOSD();
	for (;;)
	{

		DisplayMenu(tapeMenuTxt);
		SetCursor(MENU_CURSOR_X + 1, MENU_CURSOR_Y + NOPTS(tapeMenuTxt) + 2);
		PrintString("Current inputfile");
		SetCursor(MENU_CURSOR_X + 2, MENU_CURSOR_Y + NOPTS(tapeMenuTxt) + 3);
		PrintString(getBaseName(TapeInputFile));
		SetCursor(MENU_CURSOR_X + 1, MENU_CURSOR_Y + NOPTS(tapeMenuTxt) + 5);
		PrintString("Current outputfile");
		SetCursor(MENU_CURSOR_X + 2, MENU_CURSOR_Y + NOPTS(tapeMenuTxt) + 6);
		PrintString(getBaseName(TapeOutputFile));
	  again:
		switch (BrowseList(&item, NOPTS(tapeMenuTxt), NOPTS(tapeMenuTxt), NULL))
		{

			case XK_Return:

				switch(item)
				{
					case 0:
						(void)InsertTape("Input Tape", TAPE_READ);
						break;
					case 1:
						(void)InsertTape("Ouput Tape", TAPE_WRITE);
						break;
					case 2:
						if (!TapeInputFile)
						{
							goto again;
						}
						BrowseTape();
						break;
				}
				break;
			case ESCAPE:
				goto quit;
				break;
		}
	}

  quit:
	LeaveOSD(oldInDialog);
}

static char *
getBaseName(char *filename)
{
	char *s;
	if (filename)
	{
		if ((s = strrchr(filename, '/')))
		{
			return ++s;
		}
		else
		{
			return filename;
		}
	}
	return ("-none-");
}

