/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996, 1997
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Authors: Des Herriott
 *          Erik Kunze
 *
 * changed by EKU
 *******************************************************************************/
#ifndef lint
static char rcsid[] = "$Id: loadsave.c,v 3.8 1997/03/16 08:18:42 erik Rel $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include "z80.h"
#include "debug.h"
#include "resource.h"
#include "mem.h"
#include "util.h"
#include "dialog.h"
#include "loadsave.h"


#if defined(sun) && !defined(SVR4)
#define NO_FPOS_T
#endif
#define PAGE_SIZE			17
typedef struct _tapeRecord {
	uns8 id;
	uns8 type;
	char name [11];
	uns16 length;
#ifdef NO_FPOS_T
	long pos;
#else
	fpos_t pos;
#endif
	char current;
} tapeRecord;

static int getTapeContens(void);
static void displayBlocks(int);

char *TapeInputFile = NULL;
char *TapeOutputFile = NULL;

static FILE	*in = NULL;
static FILE *out = NULL;
static int numBlocks = 0;
static llist *blockList = NULL;

void 
SetLoadSaveTraps(void)
{
	int	basicRom;
	uns16 loadAddr, saveAddr;
	uns8 loadByte, saveByte;
	basicRom = GET_BASIC_ROM(GETCFG(machine));
	switch(GETCFG(machine))
	{
		default:
		case 48:
			loadAddr = LOAD_ADDR;
			loadByte = LOAD_BPT_BYTE;
			saveAddr = SAVE_ADDR;
			saveByte = SAVE_BPT_BYTE;
			break;
		case 128:
			loadAddr = LOAD_ADDR;
			loadByte = LOAD_BPT_BYTE;
			saveAddr = SAVE_ADDR;
			saveByte = SAVE_BPT_BYTE;
			break;
#ifdef XZX_PLUS3
		case 3:
			loadAddr = LOAD_ADDR_3;
			loadByte = LOAD_BPT_BYTE;
			saveAddr = SAVE_ADDR_3;
			saveByte = SAVE_BPT_BYTE;
			break;
#endif
	}
	if (RealMemory[basicRom][loadAddr] == loadByte)
	{
		RealMemory[basicRom][loadAddr] = 0xed;
		RealMemory[basicRom][loadAddr + 1] = LOAD_BPT;
	}
	else
	{
		Msg(M_WARN, "modified ROM, LOAD disabled");
	}
	if (RealMemory[basicRom][saveAddr] == saveByte)
	{
		RealMemory[basicRom][saveAddr] = 0xed;
		RealMemory[basicRom][saveAddr + 1] = SAVE_BPT;
	}
	else
	{
		Msg(M_WARN, "modified ROM, SAVE disabled");
	}
}


void
Loader(void)
{
	uns16 length, remain;
	int byte, id, crc, error = 0;
	ScreenRefresh();
#ifdef DEBUG
	if (GETCFG(debug) & D_TAPE)
	{
		Msg(M_DEBUG, "%s request: id = %02x, addr = %04X, length = %5d",
			C_FLAG ? "load" : "verify", A, IX, DE);
	}
#endif
  newtape:
	if (!in)
	{
		if (InsertTape("Load Tape", TAPE_READ) == -1)
		{
			PC = 0x0552;
			return;
		}
	}
	if (!(length = getc(in) + (getc(in) << 8)))
	{

#ifdef DEBUG
		if (GETCFG(debug) & D_TAPE)
		{
			Msg(M_DEBUG, "%s: id = ff, pos = %6d, block =     0",
				C_FLAG ? "loading" : "verifying", (int)ftell(in) - 3);
		}
#endif
		RET();
		return;
	}
	id = getc(in);
	remain = length - 1;
	length -= 2;
	if (feof(in))
	{
#ifdef DEBUG
		if (GETCFG(debug) & D_TAPE)
		{
			Msg(M_DEBUG, "end of tape reached");
		}
#endif
		(void)fclose(in);
		in = NULL;
		free(TapeInputFile);
		TapeInputFile = NULL;
		goto newtape;
	}
	if (id == A)
	{
#ifdef DEBUG
		if (GETCFG(debug) & D_TAPE)
		{
			Msg(M_DEBUG, "%s: id = %02x, pos = %6d, block = %5d",
				C_FLAG ? "loading" : "verifying", id, (int)ftell(in) - 3,
				length);
		}
#endif
		if (DE > length)
		{
			DE = length;
			error = 1;
		}
		crc = id;
		while (DE--)
		{
			byte = getc(in);
			remain--;
			if (C_FLAG)
			{
				WR_BYTE(IX, byte);
			}
			else
			{
				if (byte != RD_BYTE(IX))
				{
					error = 1;
				}
			}
			IX++;
			crc ^= byte;
		}
		if (!error)
		{
			byte = getc(in);
			remain--;
			if ((crc ^= byte) != 0)
			{
#ifdef DEBUG
				if (GETCFG(debug) & D_TAPE)
				{
					Msg(M_DEBUG, "checksum error");
				}
#endif
				error = 1;
			}
		}
	}
	else
	{
		error = 1;
	}
	if (remain)
	{

		(void)fseek(in, (long)remain, SEEK_CUR);
	}
	Z_FLAG = !error;
	C_FLAG = !error;
	IX += DE;
	DE = 0;
	RET();
}

void
Saver(void)
{
	uns16 length;
	int byte, crc;
	ScreenRefresh();
#ifdef DEBUG
	if (GETCFG(debug) & D_TAPE)
	{
		Msg(M_DEBUG, "%stape saving: addr = %04X, length = %04X",
			A ? "headerless " : "",	IX, DE);
	}
#endif
	if (!out)
	{
		if (InsertTape("Save Tape", TAPE_WRITE) == -1)
		{
			PC = 0x0552;
			return;
		}
	}
	length = DE + 2;
	(void)putc(length & 255, out);
	(void)putc(length / 256, out);
	(void)putc(A, out);
	crc = A;
	while (DE--)
	{
		byte = RD_BYTE(IX);
		(void)putc(byte, out);
		IX++;
		crc ^= byte;
	}
	(void)putc(crc, out);
	(void)fflush(out);
	SET(C_FLAG);
	IX += DE;
	DE = 0;
	RET();
}

int
InsertTape(char *title, int rdwr)
{
	char *fileName;
	FILE *fp = NULL;
	if ((fileName = FileSelector(title, rdwr == TAPE_READ, "tap", NULL, NULL)))
	{
		if (rdwr == TAPE_READ)
		{
			if (!(fp = OpenXzxFile(fileName, "r")))
			{
				Msg(M_PERR, "couldn't open <%s> for reading", fileName);
			}
			else
			{
				if (in)
				{
					(void)fclose(in);
					assert(TapeInputFile);
					free(TapeInputFile);
				}
				in = fp;
				TapeInputFile = fileName;
			}
		}
		else
		{
			if (!(fp = OpenXzxFile(fileName, "w")))
			{
				Msg(M_PERR, "couldn't open <%s> for writing", fileName);
			}
			else
			{
				if (out)
				{
					(void)fclose(out);
					assert(TapeOutputFile);
					free(TapeOutputFile);
				}
				out = fp;
				TapeOutputFile = fileName;
			}
		}
	}
	return (fp ? 0 : -1);
}

void
BrowseTape(void)
{
	int item;
	int oldInDialog;
	tapeRecord *tr;
	assert(in);

	oldInDialog = EnterOSD();

	{
		char *selectorText[PAGE_SIZE + 2];
		int i;
		selectorText[0] = "Browse Inputfile";
		for (i = 1; i <= NOPTS(selectorText); i++)
		{
			selectorText[i] = " ";
		}
		selectorText[i] = NULL;
		DisplayMenu(selectorText);
	}

	if ((item = getTapeContens()) == -1 || !numBlocks)
	{
		goto quit;
	}
	displayBlocks(item / PAGE_SIZE);
	for (;;)
	{
		switch (BrowseList(&item, numBlocks, PAGE_SIZE, displayBlocks))
		{

			case XK_Return:

				tr = (tapeRecord *)RetrieveElemList(blockList, item);
				assert(tr);
				if (!tr->current)
				{
#ifdef NO_FPOS_T
					if (fseek(in, tr->pos, SEEK_SET) == -1)
#else
					if (fsetpos(in, &(tr->pos)))
#endif
					{
						Msg(M_PERR, "couldn't set file position");
					}
				}
				goto quit;
				break;
			case XK_Escape:
				goto quit;
				break;
		}
	}
  quit:
	LeaveOSD(oldInDialog);
	DestroyList(blockList, NULL);
	blockList = NULL;
	numBlocks = 0;
}

static int
getTapeContens(void)
{
#ifdef NO_FPOS_T
	long curpos;
#else
	fpos_t curpos;
#endif
	tapeRecord *tr;
	int i;
	int current = -1;

#ifdef NO_FPOS_T
	if ((curpos = ftell(in)) == -1)
#else
	if (fgetpos(in, &curpos))
#endif
	{
		Msg(M_PERR, "couldn't get file position");
		return -1;
	}
	if (fseek(in, 0L, SEEK_SET) == -1)
	{
		Msg(M_PERR, "couldn't set file position");
		return -1;
	}
	for (;;)
	{
		tr = Malloc(sizeof(tapeRecord), "getTapeContens");
#ifdef NO_FPOS_T
		tr->pos = ftell(in);
#else
		(void)fgetpos(in, &tr->pos);
#endif
		if (!(tr->length = getc(in) + (getc(in) << 8)))
		{

			tr->id = 0xff;
		}
		else
		{

			tr->length -= 2;
			tr->id = getc(in);
		}
		if ((tr->current = (curpos == tr->pos)))
		{
			current = numBlocks;
		}
		blockList = AppendElemList(blockList, (void *)tr);
		numBlocks++;
		if (feof(in))
		{
			break;
		}
#ifdef DEBUG
		if (GETCFG(debug) & D_TAPE)
		{
			Msg(M_DEBUG, "scaning: id = %02x, pos = %6d, block = %5d",
				tr->id, tr->pos, tr->length);
		}
#endif
		if (!tr->id)
		{
			tr->type = getc(in);
			for (i = 0; i < 10; i++)
			{
				tr->name[i] = getc(in);
			}
			tr->name[10] = '\0';
			if (fseek(in, (long)(tr->length - 10), SEEK_CUR) == 1)
			{
				break;
			}
		}
		else
		{
			tr->name[0] = '\0';
			if (tr->length && fseek(in, (long)(tr->length + 1), SEEK_CUR) == -1)
			{
				break;
			}
		}
	}

	for (i = numBlocks - 2; i >= 0; i--)
	{
		if (((tapeRecord *)RetrieveElemList(blockList, i))->length)
		{
			break;
		}
	}
	numBlocks = i + 2;
	if (current == -1 || current >= numBlocks)
	{
		current = numBlocks - 1;
	}
#ifdef DEBUG
	if (GETCFG(debug) & D_TAPE)
	{
		Msg(M_DEBUG, "scaning: current = %d, total = %d", current,
			numBlocks - 1);
	}
#endif

#ifdef NO_FPOS_T
	if (fseek(in, curpos, SEEK_SET) == -1)
#else
	if (fsetpos(in, &curpos))
#endif
	{
		Msg(M_PERR, "couldn't set file position");
		return -1;
	}
	return current;
}

static void
displayBlocks(int page)
{
	int cnt, cursor_y = MENU_CURSOR_Y;
	tapeRecord *tr;
	char buf[COLS + 1];
	uns8 attr= GetAttr();
	SetAttr(MENU_SELECT_ATTR);
	SetCursor(MENU_CURSOR_X + 1, cursor_y);
	DrawBox(COLS - 2 * (MENU_CURSOR_X + 1), PAGE_SIZE);

	for (cnt = page * PAGE_SIZE;
		 cnt < (page + 1) * PAGE_SIZE && cnt < numBlocks;
		 cnt++)
	{
		tr = (tapeRecord *)RetrieveElemList(blockList, cnt);
		assert(tr);
		if (cnt == numBlocks - 1)
		{
			(void)sprintf(buf, "     *** End Of Tape ***");
		}
		else
		{
			(void)sprintf(buf, "   %03d : %10s  %5d", cnt, tr->name, tr->length);
		}
		if (tr->current)
		{
			SetAttr(MENU_DISABLED_ATTR);
			PrintString(buf);
			SetAttr(MENU_SELECT_ATTR);
		}
		else
		{
			PrintString(buf);
		}
		SetCursor(MENU_CURSOR_X + 1, ++cursor_y);
	}

	SetAttr(attr);
}

