/********************************************************************************
 * Copyright (c) Des Herriott 1993, 1994
 *               Erik Kunze   1995, 1996
 *
 * Permission to use, distribute, and sell this software and its documentation
 * for any purpose is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that the name
 * of the copyright holder not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.  The
 * copyright holder makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or implied
 * warranty. THE CODE MAY NOT BE MODIFIED OR REUSED WITHOUT PERMISSION!
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Des Herriott
 *         Erik Kunze
 *
 * changed by EKU
 *******************************************************************************/
#ifndef lint
static char z80_c[] = "$Id: z80.c,v 3.5 1997/04/14 17:49:39 erik Rel $";
#endif

#include "z80.h"
#include "debug.h"
#include "mem.h"
#include "io.h"
#include "util.h"
#ifdef XBELL_AUDIO
#include "audio.h"
#endif
#ifdef XZX_IF1
#include "if1.h"
#endif
#include "loadsave.h"
#include "monitor.h"
#include "snapshot.h"


#ifdef DEC
#undef DEC
#endif
#define PARITY(reg)		parityTbl[(reg)]

#ifndef HAVE_ENOUGH_SWAP
static void z80_DecodeCB(uns8);
static void z80_DecodeED(uns8);
static void z80_DecodeDD(uns8);
static void z80_DecodeFD(uns8);
#endif

struct z80 Cpu;
int	CarryFlag;
int AddSubFlag;
int ParityOverflowFlag;
int HalfCarryFlag;
int ZeroFlag;
int SignFlag;
int Tstates;

static uns32 work32;
static uns16 work16;
static uns8	work8, work8_2;
static int idx;
static uns8 carry, add, sub;

static int parityTbl[] = {
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1, 
};

static int halfCarryTbl[]    = { 0, 0, 1, 0, 1, 0, 1, 1 };
static int subHalfCarryTbl[] = { 0, 1, 1, 1, 0, 0, 0, 1 };

static int overflowTbl[]     = { 0, 1, 0, 0, 0, 0, 1, 0 };
static int subOverflowTbl[]  = { 0, 0, 0, 1, 1, 0, 0, 0 };

#define INC(reg) \
	{															\
		++(reg);												\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		H_FLAG = (((reg) & 0x0f) == 0);							\
		V_FLAG = ((reg) == 0x80);								\
		CLR(N_FLAG);											\
	}
#define DEC(reg) \
	{															\
		--(reg);												\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		H_FLAG = (((reg) & 0x0f) == 0x0f);						\
		V_FLAG = ((reg) == 0x7f);								\
		SET(N_FLAG);											\
	}
#define ADD(val) \
	{															\
		work16 = A + (val);										\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = halfCarryTbl[idx & 0x7];						\
		V_FLAG = overflowTbl[idx >> 4];							\
		CLR(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}
#define ADC(val) \
	{															\
		work16 = A + (val) + C_FLAG;							\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = halfCarryTbl[idx & 0x7];						\
		V_FLAG = overflowTbl[idx >> 4];							\
		CLR(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}
#define SUB(val) \
	{															\
		work16 = A - (val);										\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}
#define SBC(val) \
	{															\
		work16 = A - (val) - C_FLAG;							\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		A = work16 & 0xff;										\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}
#define AND(val) \
	{															\
		A &= (val);												\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		SET(H_FLAG);											\
		P_FLAG = PARITY(A);										\
		CLR(N_FLAG);											\
		CLR(C_FLAG);											\
	}
#define XOR(val) \
	{															\
		A ^= (val);												\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		CLR(H_FLAG);											\
		P_FLAG = PARITY(A);										\
		CLR(N_FLAG);											\
		CLR(C_FLAG);											\
	}
#define OR(val) \
	{															\
		A |= (val);												\
		S_FLAG = ((A & 0x80) != 0);								\
		Z_FLAG = (A == 0);										\
		CLR(H_FLAG);											\
		P_FLAG = PARITY(A);										\
		CLR(N_FLAG);											\
		CLR(C_FLAG);											\
	}
#define CMP(val) \
	{															\
		work16 = A - (val);										\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		S_FLAG = ((work16 & 0x80) != 0);						\
		Z_FLAG = ((work16 & 0xff) == 0);						\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work16 & 0x0100) != 0);						\
	}

#define _CMP_(val) \
	{															\
		work16 = A - (val);										\
		idx = ((A & 0x88) >> 1) |								\
			  (((val) & 0x88) >> 2) |							\
			  ((work16 & 0x88) >> 3);							\
		S_FLAG = ((work16 & 0x80) != 0);						\
		Z_FLAG = ((work16 & 0xff) == 0);						\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		SET(N_FLAG);											\
	}
#define ADD_HL(reg) \
	{															\
		work32 = HL + (reg);									\
		idx = ((HL & 0x0800) >> 9) |							\
			  (((reg) & 0x0800) >> 10) |						\
			  ((work32 & 0x0800) >> 11);						\
		HL = work32 & 0xffff;									\
		H_FLAG = halfCarryTbl[idx];								\
		CLR(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}
#define ADC_HL(reg) \
	{															\
		work32 = HL + (reg) + C_FLAG;							\
		idx = ((HL & 0x8800) >> 9) |							\
			  (((reg) & 0x8800) >> 10) |						\
			  ((work32 & 0x8800) >> 11);						\
		HL = work32 & 0xffff;									\
		S_FLAG = ((HL & 0x8000) != 0);							\
		Z_FLAG = (HL == 0);										\
		H_FLAG = halfCarryTbl[idx & 0x7];						\
		V_FLAG = overflowTbl[idx >> 4];							\
		CLR(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}
#define SBC_HL(reg) \
	{															\
		work32 = HL - (reg) - C_FLAG;							\
		idx = ((HL & 0x8800) >> 9) |							\
			  (((reg) & 0x8800) >> 10) |						\
			  ((work32 & 0x8800) >> 11);						\
		HL = work32 & 0xffff;									\
		S_FLAG = ((HL & 0x8000) != 0);							\
		Z_FLAG = (HL == 0);										\
		H_FLAG = subHalfCarryTbl[idx & 0x7];					\
		V_FLAG = subOverflowTbl[idx >> 4];						\
		SET(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}
#define ADD_XX(reg) \
	{															\
		work32 = XX + (reg);									\
		idx = ((XX & 0x0800) >> 9) |							\
			  (((reg) & 0x0800) >> 10) |						\
			  ((work32 & 0x0800) >> 11);						\
		XX = work32 & 0xffff;									\
		H_FLAG = halfCarryTbl[idx];								\
		CLR(N_FLAG);											\
		C_FLAG = ((work32 & 0x10000) != 0);						\
	}
#define RLC(reg) \
	{															\
		C_FLAG = (((reg) & 0x80) != 0);							\
		(reg) = ((reg) << 1) | C_FLAG;							\
		_SHIFT_TEST_((reg));									\
	}
#define RRC(reg) \
	{															\
		C_FLAG = (reg) & 0x01;									\
		(reg) = ((reg) >> 1) | (C_FLAG << 7);					\
		_SHIFT_TEST_((reg));									\
	}
#define RL(reg) \
	{															\
		work8_2 = (((reg) & 0x80) != 0);						\
		(reg) = ((reg) << 1) | C_FLAG;							\
		C_FLAG = work8_2;										\
		_SHIFT_TEST_((reg));									\
	}
#define RR(reg) \
	{															\
		work8_2 = (reg) & 0x01;									\
		(reg) = ((reg) >> 1) | (C_FLAG << 7);					\
		C_FLAG = work8_2;										\
		_SHIFT_TEST_((reg));									\
	}
#define SLA(reg) \
	{															\
		C_FLAG = (((reg) & 0x80) != 0);							\
		(reg) <<= 1;											\
		_SHIFT_TEST_((reg));									\
	}
#define SRA(reg) \
	{															\
		C_FLAG = (reg) & 0x01;									\
		work8_2 = (reg) & 0x80;									\
		(reg) = ((reg) >> 1) | work8_2;							\
		_SHIFT_TEST_((reg));									\
	}
#define SLI(reg) \
	{															\
		C_FLAG = (((reg) & 0x80) != 0);							\
		(reg) = ((reg) << 1) | 0x01;							\
		_SHIFT_TEST_((reg));									\
	}
#define SRL(reg) \
	{															\
		C_FLAG = (reg) & 0x01;									\
		(reg) >>= 1;;											\
		_SHIFT_TEST_((reg));									\
	}
#define _SHIFT_TEST_(reg) \
	{															\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		CLR(H_FLAG);											\
		P_FLAG = PARITY((reg));									\
		CLR(N_FLAG);											\
	}
#define BIT(bit, reg) \
	{															\
		Z_FLAG = (((reg) & (0x01 << (bit))) == 0);				\
		SET(H_FLAG);											\
		P_FLAG = Z_FLAG;										\
		CLR(N_FLAG);											\
		S_FLAG = (((reg) & (0x01 << (bit))) == 0x80);			\
	}
#define IN(port, reg) \
	{															\
		(reg) = InByte((port));									\
		S_FLAG = (((reg) & 0x80) != 0);							\
		Z_FLAG = ((reg) == 0);									\
		CLR(H_FLAG);											\
		P_FLAG = PARITY((reg));									\
		CLR(N_FLAG);											\
	}
#define CALL \
	{															\
		PUSH(PC + 2);											\
		PC = RD_WORD(PC);										\
	}
#define RST(addr) \
	{															\
		PUSH(PC);												\
		PC = (addr);											\
	}
#ifdef DEBUG
#define Z80_INTERRUPT \
	{															\
		if (GETCFG(debug) & D_INTERRUPT)						\
		{														\
			Msg(M_DEBUG,										\
				"interrupt: IFF1=%d, IFF2=%d", IFF1, IFF2);		\
		}														\
																\
		if (IFF1)												\
		{														\
									\
			if (RD_BYTE(PC) == 0x76)							\
			{													\
				PC++;											\
			}													\
																\
			IFF1 = 0;											\
			IFF2 = 0;											\
			PUSH(PC);											\
																\
			if (IM == 2)										\
			{													\
				Tstates += 19;									\
				PC = RD_WORD((I << 8) | 0xff);					\
			}													\
			else												\
			{													\
				Tstates += 13;									\
				PC = 0x0038;									\
			}													\
		}														\
	}
#else
#define Z80_INTERRUPT \
	{															\
		if (IFF1)												\
		{														\
									\
			if (RD_BYTE(PC) == 0x76)							\
			{													\
				PC++;											\
			}													\
																\
			IFF1 = 0;											\
			IFF2 = 0;											\
			PUSH(PC);											\
																\
			if (IM == 2)										\
			{													\
				Tstates += 19;									\
				PC = RD_WORD((I << 8) | 0xff);					\
			}													\
			else												\
			{													\
				Tstates += 13;									\
				PC = 0x0038;									\
			}													\
		}														\
	}
#endif
#define Z80_M1(op) \
	{															\
		(op) = RD_BYTE(PC++);									\
																\
			\
																\
		R = (R & 0x80) | (++R & 0x7f);							\
	}

void
Z80_Init(void)
{
	Tstates = 0;
	S_FLAG = 0;
	Z_FLAG = 0;
	H_FLAG = 0;
	P_FLAG = 0;
	N_FLAG = 0;
	C_FLAG = 0;
	Z80_Reset();
}

void
Z80_StoreFlags(void)
{
	if (S_FLAG) F |= F_NEG;      else F &= ~F_NEG;
	if (Z_FLAG) F |= F_ZERO;     else F &= ~F_ZERO;
	if (H_FLAG) F |= F_HCARRY;   else F &= ~F_HCARRY;
	if (P_FLAG) F |= F_OVERFLOW; else F &= ~F_OVERFLOW;
	if (N_FLAG) F |= F_NADD;     else F &= ~F_NADD;
	if (C_FLAG) F |= F_CARRY;    else F &= ~F_CARRY;
}

void
Z80_RetrieveFlags(void)
{
	S_FLAG = F & F_NEG      ? 1 : 0;
	Z_FLAG = F & F_ZERO     ? 1 : 0;
	H_FLAG = F & F_HCARRY   ? 1 : 0;
	P_FLAG = F & F_OVERFLOW ? 1 : 0;
	N_FLAG = F & F_NADD     ? 1 : 0;
	C_FLAG = F & F_CARRY    ? 1 : 0;
}

void
Z80_Reset(void)
{
	PC   = 0;
	I    = 0;
	R    = 0;
	IM   = 0;
	IFF1 = 0;
	IFF2 = 0;
	IoReset();
}

void
Z80_NMI(void)
{

	Tstates += 15;
	IFF2 = IFF1;
	IFF1 = 0;
	if (RD_BYTE(PC) == 0x76)
	{
		PC++;
	}
	PUSH(PC);
	PC = 0x0066;
#ifdef XZX_MF128

	if (GETCFG(mf128_active))
	{
		(void)InByte(P_MF128IN);
	}
#endif
}

#ifndef HAVE_ENOUGH_SWAP
static void
z80_DecodeCB(uns8 op)
{
	switch(op)
	{
#		include "z80_cbops.c"
	}
}
static void
z80_DecodeED(uns8 op)
{
	switch(op)
	{
#		include "z80_edops.c"
	}
}
static void
z80_DecodeDD(uns8 op)
{
	switch(op)
	{
#		define XX  IX
#		define LXX LIX
#		define HXX HIX
#		include "z80_xxops.c"
#		undef XX
#		undef LXX
#		undef HXX
	}
}
static void
z80_DecodeFD(uns8 op)
{
	switch(op)
	{
#		define NO_RCSID
#		define XX  IY
#		define LXX LIY
#		define HXX HIY
#		include "z80_xxops.c"
#		undef XX
#		undef LXX
#		undef HXX
#		undef NO_RCSID
	}
}
#endif

